from typing import Final

KEY: Final[str] = "_key"
"""The key identifier used in the query."""

ID: Final[str] = "_id"
"""The ID identifier used in the query."""

REV: Final[str] = "_rev"
"""The revision identifier used in the query."""

FROM: Final[str] = "_from"
"""The 'from' field used in query relations."""

TO: Final[str] = "_to"
"""The 'to' field used in query relations."""

DYNAMIC_ALIAS: Final[str] = "DynamicAlias"
"""Alias used for dynamic query operations."""
